/*
Copyright (c) 2004 EarthLink, Inc.
All rights reserved.

FPeer Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions 
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package net.earthlink.research.p2p;

import gov.nist.javax.sip.Utils;

import java.io.ByteArrayInputStream;
import java.io.IOException;

import java.text.ParseException;

import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.TooManyListenersException;
import java.util.Vector;

import javax.sip.ClientTransaction;
import javax.sip.Dialog;
import javax.sip.InvalidArgumentException;
import javax.sip.ListeningPoint;
import javax.sip.ObjectInUseException;
import javax.sip.PeerUnavailableException;
import javax.sip.RequestEvent;
import javax.sip.ResponseEvent;
import javax.sip.ServerTransaction;
import javax.sip.SipException;
import javax.sip.SipFactory;
import javax.sip.SipListener;
import javax.sip.SipProvider;
import javax.sip.SipStack;
import javax.sip.Transaction;
import javax.sip.TransactionAlreadyExistsException;
import javax.sip.TransactionUnavailableException;
import javax.sip.TransportNotSupportedException;
import javax.sip.address.Address;
import javax.sip.address.AddressFactory;
import javax.sip.address.SipURI;
import javax.sip.header.AcceptHeader;
import javax.sip.header.AllowHeader;
import javax.sip.header.CSeqHeader;
import javax.sip.header.CallIdHeader;
import javax.sip.header.ContactHeader;
import javax.sip.header.ContentTypeHeader;
import javax.sip.header.EventHeader;
import javax.sip.header.ExpiresHeader;
import javax.sip.header.FromHeader;
import javax.sip.header.HeaderFactory;
import javax.sip.header.MaxForwardsHeader;
import javax.sip.header.SubscriptionStateHeader;
import javax.sip.header.ToHeader;
import javax.sip.header.ViaHeader;
import javax.sip.message.MessageFactory;
import javax.sip.message.Request;
import javax.sip.message.Response;

import net.earthlink.research.fsp.FSPclient;
import net.earthlink.research.fsp.fspd;

import net.earthlink.research.p2p.ui.WindowManager;

import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.FileAppender;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.PatternLayout;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.output.Format;
import org.jdom.input.SAXBuilder;
import org.jdom.output.XMLOutputter;

/**
 * This class is FPeer, the standalone client in SIP Peer File Exchange (SPFX). 
 *
 *@author Mark Petrovic
 */
public class FPeer implements SipListener {

   private          int               oneDay                       = 86400;
   private          int               xferServerPort;     
   private          int               minNetworkSize;
   
   private          SipProvider       sipProvider;
   private          AddressFactory    addressFactory;
   private          MessageFactory    messageFactory;
   private          HeaderFactory     headerFactory;
   private          SipStack          sipStack;
   private          SipURI            mySIPURI;
   private          SipURI            myContactURI;
   private          Address           myFromNameAddress;
   private          Address           myContactAddress;
   private          FromHeader        myFromHeader;
   private          ContactHeader     myContactHeader;
   private          fspd              fspServer;      
   private          WindowManager     windowManager;

   private          String            version                      = "v0.14.9";
   private          String            windowTitle;
   private          String            transportProto;
   private          String            transportPort;
   private          String            myFromName;
   private          String            myStackAddress;
   private          String            myFromSIPAddress;
   private          String            myDisplayName;
   private          String            peerListMIMEType             = "x-earthlink-peerlist-fpx";
   private          String            fileListMIMEType             = "x-earthlink-files-fpx";
   private          String            cacheHitMIMEType             = "x-earthlink-cachehits-fpx";
   private          String            searchFilterDocumentMIMEType = "x-earthlink-searchfilter-fpx";
   private          String            shareDirectory;

   final   public   Logger            logger;   
   public  static   FPeer             fpeer;
   public           RunTimeProperties rtProps;
   public           ShareList         shareList;   
   public  volatile NetherHosts       netherHosts;
   public  volatile Subscriptions     subscriptions;
   public  volatile KnownHosts        knownHosts;
   public  volatile Subscribers       subscribers;
  
   public           MessageID         messageIDs;
   
   //-------------------------------------------------------------------

   public FPeer(String[] argv) {
      SipFactory      sipFactory;
      ListeningPoint  lp;
      PatternLayout   layout;
      ConsoleAppender appender;
      FileAppender    fileAppender;

      fpeer         = this;
      rtProps       = new RunTimeProperties(argv);

      // Logger stuff

      logger   = Logger.getLogger(this.getClass().getPackage().getName());
      layout   = new PatternLayout("%d{dd MMM yyyy HH:mm:ss}, %l, %m %n");
      appender = new ConsoleAppender(layout);
      logger.addAppender(appender);
      /*
      try {
         String fname = rtProps.getProperty("javax.sip.STACK_NAME") + ".debug";
         fileAppender = new FileAppender(layout, fname);
         logger.addAppender(fileAppender);
      } catch (java.io.IOException e) {
         e.printStackTrace();
      }
      */

      if( rtProps.getProperty("net.earthlink.research.p2p.loglevel").equalsIgnoreCase("debug") ) {
         logger.setLevel((Level ) Level.DEBUG);
      } else if( rtProps.getProperty("net.earthlink.research.p2p.loglevel").equalsIgnoreCase("info") ) {
         logger.setLevel((Level ) Level.INFO);
      } else if( rtProps.getProperty("net.earthlink.research.p2p.loglevel").equalsIgnoreCase("warn") ) {
         logger.setLevel((Level ) Level.WARN);
      } else if( rtProps.getProperty("net.earthlink.research.p2p.loglevel").equalsIgnoreCase("error") ) {
         logger.setLevel((Level ) Level.ERROR);
      } else if( rtProps.getProperty("net.earthlink.research.p2p.loglevel").equalsIgnoreCase("fatal") ) {
         logger.setLevel((Level ) Level.FATAL);
      } else if( rtProps.getProperty("net.earthlink.research.p2p.loglevel").equalsIgnoreCase("all") ) {
         logger.setLevel((Level ) Level.ALL);
      } else {
         logger.setLevel((Level ) Level.WARN);
      }

      /* A netherHost is a host known to us, but to which we are not
      subscribed.  If we need to subscribe to more hosts to expand
      our network, we use hosts on netherHosts.  They are 'spares'. */
      netherHosts   = new NetherHosts();

      /* Hosts to which we are subscribed, and which are subscribed to
      us, respectively */
      subscriptions = new Subscriptions();
      subscribers   = new Subscribers();

      /* knownHosts is the set of all hosts known to us; the union of
      netherHosts and subscriptions */
      knownHosts    = new KnownHosts();

      /* The list of call-ids or equivalent that prevents us from
      forwarding a search request more than once */
      messageIDs    = new MessageID();
            
      //----------- File share list

      shareDirectory = rtProps.getProperty("net.earthlink.research.p2p.sharedDir", "./share");
      if( ! shareDirectory.endsWith("/")) 
         shareDirectory = shareDirectory + "/";   
      shareList      = new ShareList(shareDirectory);
      
      //--------- Create personal headers, etc.

      myStackAddress   = rtProps.getProperty("javax.sip.IP_ADDRESS", "127.0.0.1");
      myDisplayName    = rtProps.getProperty("net.earthlink.research.p2p.myDisplayName", "H. Munster");
      myFromSIPAddress = myStackAddress;
      myFromName       = rtProps.getProperty("net.earthlink.research.p2p.myFromName", "misterNobody");
      
      xferServerPort   = Integer.parseInt(
                            rtProps.getProperty("net.earthlink.research.p2p.xferServerPort", "2121")
                         );

      windowTitle      = rtProps.getProperty("javax.sip.STACK_NAME") + "/" + version;
      windowManager    = new WindowManager( windowTitle );

      minNetworkSize   = Integer.parseInt( 
                            rtProps.getProperty("net.earthlink.research.p2p.minNetworkSize", "1")
                         );

      try {

         //		----------- SIP stack stuff

         sipFactory = SipFactory.getInstance();
         sipFactory.setPathName("gov.nist");

         sipStack = sipFactory.createSipStack(rtProps);

         headerFactory     = sipFactory.createHeaderFactory();
         addressFactory    = sipFactory.createAddressFactory();
         messageFactory    = sipFactory.createMessageFactory();

         transportPort     = rtProps.getProperty("net.earthlink.research.p2p.listeningpoint.PORT", "5060");
         transportProto    = rtProps.getProperty("net.earthlink.research.p2p.listeningpoint.PROTO", "udp");

         lp                = sipStack.createListeningPoint(Integer.parseInt(transportPort), transportProto);
         sipProvider       = sipStack.createSipProvider(lp);

         addressFactory    = sipFactory.createAddressFactory();
         headerFactory     = sipFactory.createHeaderFactory();
         messageFactory    = sipFactory.createMessageFactory();

         mySIPURI          = addressFactory.createSipURI(myFromName, myFromSIPAddress);

         myFromNameAddress = addressFactory.createAddress(mySIPURI);
         myFromNameAddress.setDisplayName(myDisplayName);

         myFromHeader      = headerFactory.createFromHeader(myFromNameAddress, null);

         myContactURI      = addressFactory.createSipURI(myFromName, sipStack.getIPAddress());
         myContactURI.setPort(sipProvider.getListeningPoint().getPort());
         myContactURI.setTransportParam(sipProvider.getListeningPoint().getTransport());

         myContactAddress  = addressFactory.createAddress(myContactURI);
         myContactAddress.setDisplayName(myDisplayName);
         myContactHeader   = headerFactory.createContactHeader(myContactAddress);

      } catch (PeerUnavailableException e) {
         e.printStackTrace();
         System.err.println(e.getMessage());
         if (e.getCause() != null)
            e.getCause().printStackTrace();
         logger.error("Exiting.");
         System.exit(0);
      } catch (TransportNotSupportedException e) {
         e.printStackTrace();
         logger.error("Exiting.");
         System.exit(0);
      } catch (InvalidArgumentException e) {
         e.printStackTrace();
         logger.error("Exiting.");
         System.exit(0);
      } catch (ObjectInUseException e) {
         e.printStackTrace();
         logger.error("Exiting.");
         System.exit(0);
      } catch (ParseException e) {
         e.printStackTrace();
         logger.error("Exiting.");
         System.exit(0);
      }
   }

   private void networkStart() {
      
      // Start the fsp server... 
      
      fspServer = new fspd(myStackAddress, xferServerPort, shareDirectory, fspd.MODE_DIGEST);
      fspServer.start();
      
      // Start the SIP listener... 
      
      try {
         sipProvider.addSipListener(this);
         logger.warn("SIP listener started.");
      } catch (TooManyListenersException e) {
         e.printStackTrace();
      }

      /* Build network...
      *  
      */
      new Thread() {
         boolean spin           = true;
         int     sleepy         = 15000;
         
         public void run() {
            String host;

            setName("NetworkBuildThread");
            while ( spin ) {            
               if( (subscriptions.size() < minNetworkSize) ) {
                  for( Iterator i = netherHosts.iterator(); i.hasNext(); ) {
                     host = (String )i.next();
                     subscribeToPeerList(host);
                  } 
               }
               try {
                  sleep(sleepy);
               }
               catch (InterruptedException e) {
                  e.printStackTrace();
               }
            }
      }
      }.start();

      /* ...periodically check for addition of new hosts, and notify
         subscribers accordingly 
      */
      new Thread() {
         public void run() {
            long sleepy = 15000;
            
            setName("NewPeersNotifyThread");
            try {
               while (true) {
                  sleep(sleepy);
                  if( knownHosts.hasChanged() ) {
                     for( Iterator i = subscribers.iterator(); i.hasNext(); ) {
                        sendPeerListNotify((Subscriber )i.next());
                     }
                  }
               }
            } catch (Exception e) {
               e.printStackTrace();
            }
         
         }
      }.start();

      /* The Subscription-list heartbeat.  This verifies that
         a host to which we are subscribed is still on the network.
      */
      new Thread() {
         public void run() {
            long sleepy = 60000;
            
            setName("HeartBeatThread");
            try {
               while (true) {
                  sleep(sleepy);
                  for( Iterator i = subscriptions.iterator(); i.hasNext(); ) {
                     pingSubscriptions((Subscription )i.next());
                  }
               }
            } catch (Exception e) {
               e.printStackTrace();
            }
         
         }
      }.start();

   }

   //==============================================================
   //==============================================================
   //==============================================================

   // SipListener interface contract

   /**
   ----------------------------------------------------------------
   */
   public void processRequest(RequestEvent requestEvent) {
      StringBuffer      sb;
      String            method;
      Request           request;
      ServerTransaction serverTransactionId;

      request             = requestEvent.getRequest();
      serverTransactionId = requestEvent.getServerTransaction();
      method              = request.getMethod();

      logger.debug("received = " + request);

      if (method.equals(Request.INVITE)) {
         logger.debug("   calling processInvite()");
         processInvite(requestEvent, serverTransactionId);
         logger.debug("   returned from processInvite()");
      } else if (method.equals(Request.SUBSCRIBE)) {
         logger.debug("   calling processSubscribe()");
         processSubscribe(requestEvent, serverTransactionId);
         logger.debug("   returned from processSubscribe()");
      } else if (method.equals(Request.NOTIFY)) {
         logger.debug("   calling processNotify()");
         processNotify(requestEvent, serverTransactionId);
         logger.debug("   returned from processNotify()");
      } else if (method.equals(Request.ACK)) {
         logger.debug("   calling processAck()");
         processAck(requestEvent, serverTransactionId);
         logger.debug("   returned from processAck()");
      } else if (method.equals(Request.BYE)) {
         logger.debug("   calling processBye()");
         processBye(requestEvent, serverTransactionId);
         logger.debug("   returned from processBye()");
      }
      logger.debug("exit processRequest() = " + method);
   }

   //----------------------------------------------------------------
   public void processResponse(ResponseEvent responseReceivedEvent) {
      int             statusCode;
      int             ackCount;
      Request         ackRequest;
      Response        response;
      Transaction     tid;
      SipURI          toURI;
      SipURI          requestURI;
      SDPMessage      sdpMessage;               
      Subscription    subscription;
      String          method;
      final Dialog    dialog;
      final String    digest;
      final String    saveAs;                              
      final int       fileSize;                             
      final FSPclient fspClient;

      response   = (Response) responseReceivedEvent.getResponse();
      logger.debug("received = " + response);

      tid        = responseReceivedEvent.getClientTransaction();
      statusCode = response.getStatusCode();
      method     = ((CSeqHeader) response.getHeader(CSeqHeader.NAME)).getMethod();
      toURI      = (SipURI) ( ((ToHeader) (response.getHeader(ToHeader.NAME))).getAddress()).getURI();
      
      try {
         if (statusCode == Response.OK ) {

            if( method .equals(Request.INVITE)) {
               dialog  = tid.getDialog();                              
               
               ackRequest = dialog.createRequest(Request.ACK);
               dialog.sendAck(ackRequest);
               ((DialogApplicationData )dialog.getApplicationData()).ackCount++;
                              
               ackCount = ( (DialogApplicationData )dialog.getApplicationData()).ackCount;
               
               // No matter how many OK's we get, initiate only one FSPclient process.

               if( ackCount == 1 ) {                                
                  
                  sdpMessage = new SDPMessage(response.getRawContent());               
                  digest     = (String )toURI.getUser();
                  saveAs     = ( (DialogApplicationData )dialog.getApplicationData() ).getFileName();                              
                  fileSize   = ( (DialogApplicationData )dialog.getApplicationData() ).fileSize;                              
                  // 1st argument should really be per SDP, but lazy
                  fspClient  = new FSPclient(toURI.getHost(), sdpMessage.getMediaPort());

                  fspClient.addObserver(windowManager);
                  windowManager.addDownloadClient(fspClient, saveAs, fileSize);
                  
                  logger.warn("Save as = " + saveAs);
               
                  new Thread() {
                     public void run () {
                        SharedFile sf;

                        setName("ProcessResponseThread");
                        try {
                           fspClient.download(digest, shareDirectory, saveAs);
                           sf = new SharedFile(shareDirectory, saveAs);
                           shareList.add(sf);
                           windowManager.addSharedFile(sf);
                           logger.warn("Download complete, saved as = <shareDir>/" + saveAs);
                           sendBye(dialog);
                        }
                        catch (IOException e) {

                           /* FSP timeout or other I/O problem.  includes
                           InterruptedIOException on transfer timeout */

                           /* Remove at the filesystem level fragments
                           of saveAs if there are any.  */
                           shareList.remove(saveAs);

                           logger.error("FSP timeout or error downloading file = " + saveAs);
                           sendBye(dialog);
                        }                     
                     }                  
                  }.start();               
               }
               
            } else if ( method.equals(Request.SUBSCRIBE) ) {                              
               if( toURI.getUser().equals("peer-list") ) {
                  dialog = tid.getDialog();
                  subscription = new Subscription(toURI, response, dialog);
                  subscriptions.add(subscription);

                  // why do I need this loc??  4 sep 2004
                  subscriptions.removePending(toURI.getHost());            
               }
            } else if ( method .equals(Request.NOTIFY) ) {
               ;
            }
         }         
      } catch (SipException ex) {
         ex.printStackTrace();
         logger.error("Exiting.");
         System.exit(0);
      }
   }

   public void processTimeout(javax.sip.TimeoutEvent timeoutEvent) {
      Request     request;
      SipURI      toURI;
      Transaction transaction;
      String      method;
      
      if (timeoutEvent.isServerTransaction()) {
         transaction = timeoutEvent.getServerTransaction();
      } else {
         transaction = timeoutEvent.getClientTransaction();
      }
      
      request = transaction.getRequest();
      method  = ((CSeqHeader) request.getHeader(CSeqHeader.NAME)).getMethod();
      toURI   = (SipURI )( ((ToHeader) (request.getHeader(ToHeader.NAME))).getAddress()).getURI();
      
      if( method.equals(Request.INVITE) ) {
         ; // XXx: what to do??
      }
      else if( method.equals(Request.SUBSCRIBE) ) {
         if( toURI.getUser().equals("peer-list")) {
            subscriptions.removePending(toURI.getHost());
            logger.warn("Timeout adding host " + toURI.getHost() + " to network.");
         }
         else if (toURI.getUser().equals("ping")) {

            logger.warn("Timeout on heartbeat from host " + toURI.getHost());

            // Remove the host from the list of active subscriptions

            subscriptions.remove(toURI.getHost());

            /* If the host is not responsive to a heartbeat, chances
            are it will not be responsive to a peerList update from us.
            Remove it as a subscriber, too.  */

            subscribers.remove(toURI.getHost());
         }
         else {
            ; // drop on floor
         }
      }
      else if( method.equals(Request.NOTIFY) ) {
         subscribers.remove(toURI.getHost());
      }      
      else if (method.equals(Request.BYE)) {
         ; // XXX: what to do??
      }
      //logger.warn("Timeout:  Request = " + request);
   }

   // End SipListener interface contract

   //==============================================================
   //==============================================================
   //==============================================================

   /** Process the ACK request.
   ----------------------------------------------------------------
   */
   private void processAck( RequestEvent requestEvent, ServerTransaction serverTransaction) {
      SipProvider sipProvider;
      Dialog      dialog;      
      Request     request;

      request     = requestEvent.getRequest();

      //logger.warn("processAck Request = " + request);                  

      sipProvider = (SipProvider) requestEvent.getSource();
      dialog      = serverTransaction.getDialog();         

      logger.debug("processAck Dialog = " + dialog);                  
   }


   /** Process the Subscribe request.
   ----------------------------------------------------------------
   */
   private void processSubscribe( RequestEvent requestEvent, ServerTransaction serverTransaction) {

      long              expiresAt;
      SipProvider       sipProvider;
      Request           request;
      Response          response;
      SipURI            fromURI;
      Dialog            dialog;
      ServerTransaction st;
      ToHeader          toHeader;
      ExpiresHeader     expiresHeader;
      CallIdHeader      callIdHeader;
      ContactHeader     contactHeader;

      final Subscriber  subscriber;
      String            requestURI;
      String            regex;

      sipProvider = (SipProvider) requestEvent.getSource();
      request     = requestEvent.getRequest();
      requestURI  = request.getRequestURI().toString();
      fromURI     = (SipURI) ( ((FromHeader) (request.getHeader(FromHeader.NAME))).getAddress()).getURI();

      try {

         response = messageFactory.createResponse(Response.OK, request);

         /* 
         My reading of RFC 3261 sec 8.2.6.2 is that both stateful and
         stateless responses require the To tag to be set.

         For the stateless responses that occur during file searches,
         I have no easy way of guaranteeing that the To tag would be the
         same for a retransmitted request (8.2.7).  Work around this by
         *never* retransmitting a request, which is quite practicable.
         */

         toHeader = (ToHeader ) response.getHeader(ToHeader.NAME);
         toHeader.setTag(Utils.generateTag());

         response.setHeader(myContactHeader);

         if( (requestURI.toLowerCase()).startsWith("sip:peer-list") ) {

            expiresHeader = headerFactory.createExpiresHeader(oneDay);  
            response.setHeader(expiresHeader);

            st = requestEvent.getServerTransaction();
            if( st == null )
               st = sipProvider.getNewServerTransaction(request);
            st.sendResponse(response);

            dialog = st.getDialog();

            if( dialog != null ) {

               /* The oneDay parameter to new Subscribers constructor
                  is now a constant and essentially superfluous.  It was
                  presumed variable in earlier versions.
               */
               subscriber = new Subscriber(fromURI, response, dialog, oneDay);
               subscribers.add(subscriber);
               netherHosts.add(fromURI.getHost());

               // Send the peerList statefully w/in sendPeerListNotify()...
               new Thread() {
                  public void run() {
                     setName("SendPeerListNotifyThread");
                     sendPeerListNotify(subscriber);
                  }
               }
               .start();
            } else {
               logger.warn("Dialog null.  Subscriber not added.");
            }

         } else if ( (requestURI.toLowerCase()).startsWith("sip:ping") ) {

            // Simple response to a hearbeat ping from this subscriber 

            expiresHeader = headerFactory.createExpiresHeader(oneDay);  
            response.setHeader(expiresHeader);

            st = requestEvent.getServerTransaction();
            if( st == null )
               st = sipProvider.getNewServerTransaction(request);
            st.sendResponse(response);

            logger.warn("Heartbeat response sent to peer = " + fromURI.getHost());

         } else if ( (requestURI.toLowerCase()).startsWith("sip:search") ) {
            
            // Send response statelessly

            expiresHeader = headerFactory.createExpiresHeader(0);  
            response.setHeader(expiresHeader);

            sipProvider.sendResponse(response);

            // Notify sender of cache hits and forward message to all knownHosts
            final Request req = (Request )request.clone();
            new Thread() {
               public void run() {
                  setName("CacheHitNotifyThread");
                  sendSearchNotify(req);   
                  forward(req);   
               }
            }.start();

           
         } else {
            // send 'user unknown' or some such
         }

         logger.debug("sent = " + response);

      }
      catch (ParseException e) {
         e.printStackTrace();
         logger.error("Exiting.");
         System.exit(0);
      }
      catch (InvalidArgumentException e) {
         e.printStackTrace();
         logger.error("Exiting.");
         System.exit(0);
      }
      catch (TransactionUnavailableException e) {
         e.printStackTrace();
         logger.error("Exiting.");
         System.exit(0);
      }
      catch (SipException e) {
         e.printStackTrace();
         logger.error("Exiting.");
         System.exit(0);
      }
      
   }

   /** Process the Notify request.
   ----------------------------------------------------------------
   */
   private void processNotify( RequestEvent requestEvent, ServerTransaction serverTransaction) {
      byte[]            messageBody;
      long              expiresAt;
      SipProvider       sipProvider;
      Request           request;
      Response          response;
      ServerTransaction st;
      final SipURI      fromURI;
      ContentTypeHeader contentTypeHeader;
      ToHeader          toHeader;
      List              children;
      String            peerAddress;
      final Document    doc;

      sipProvider = (SipProvider) requestEvent.getSource();
      request     = requestEvent.getRequest();
      fromURI     = (SipURI) ( ((FromHeader) (request.getHeader(FromHeader.NAME))).getAddress()).getURI();
      st          = serverTransaction;

      try {
         messageBody = request.getRawContent();
         doc         = new SAXBuilder().build(new ByteArrayInputStream(messageBody));

         // Interpret the XML body based on MIME content sub-type

         contentTypeHeader = (ContentTypeHeader) request.getHeader(ContentTypeHeader.NAME);

         if( contentTypeHeader.getContentSubType().equals(peerListMIMEType) ) {

            response = messageFactory.createResponse(200, request);
            response.setHeader(myContactHeader);

            if( st == null )
               st = sipProvider.getNewServerTransaction(request);
            st.sendResponse(response);
            logger.debug("sent = " + response);

            children = doc.getRootElement().getChildren("peer");

            for( Iterator i = children.iterator(); i.hasNext(); ) {
               peerAddress = ( (Element )i.next() ).getText();
               netherHosts.add(peerAddress);
            }

         } else if( contentTypeHeader.getContentSubType().equals(fileListMIMEType) ) {
            ;
         } else if( contentTypeHeader.getContentSubType().equals(cacheHitMIMEType) ) {

            response = messageFactory.createResponse(200, request);
            response.setHeader(myContactHeader);

            // Need this according to sec 8.2.6.2
            toHeader = (ToHeader ) response.getHeader(ToHeader.NAME);
            toHeader.setTag(Utils.generateTag());

            sipProvider.sendResponse(response);

            logger.debug("sent = " + response);

            new Thread() {
               public void run() {
                  setName("WindowManagerUpdateThread");
                  windowManager.updateCacheHits(doc, fromURI.getHost());
               }
            }.start();
         }

      } catch (TransactionAlreadyExistsException ex) {
         //ex.printStackTrace();
      } catch (JDOMException ex) {
         logger.error("JDOMException");
         ex.printStackTrace();
      } catch (Exception ex) {
         ex.printStackTrace();
         logger.fatal("Fatal error");
         System.exit(0);
      }
   }

   /** Process the invite request.
   ----------------------------------------------------------------
   */
   private void processInvite( RequestEvent requestEvent, ServerTransaction serverTransaction) {
      byte[]            content;
      boolean           haveFile;
      SipProvider       sipProvider;
      Request           request;
      Response          response;      
      Dialog            dialog;
      ServerTransaction st;
      ContentTypeHeader contentTypeHeader;      
      String            digest;
      SDPMessage        sdpMessage;      

      request     = requestEvent.getRequest();
      digest      = ( (SipURI )request.getRequestURI() ).getUser();      
      haveFile    = shareList.isShared(digest);
      sipProvider = (SipProvider) requestEvent.getSource();
      try {
         
         if( haveFile ) {                    
            response          = messageFactory.createResponse(200, request);
            content           = request.getRawContent();                        
            sdpMessage        = new SDPMessage(content);
            contentTypeHeader = headerFactory.createContentTypeHeader("application", "sdp");
            sdpMessage.setMediaPort(xferServerPort);
            response.setContent(sdpMessage.getBytes(), contentTypeHeader);
         }
         else {
            response          = messageFactory.createResponse(Response.NOT_FOUND, request);
         }
         
         ToHeader toHeader = (ToHeader) response.getHeader(ToHeader.NAME);
         toHeader.setTag(Utils.generateTag());         
         response.addHeader(myContactHeader);

         st     = sipProvider.getNewServerTransaction(request);                  
         dialog = st.getDialog();
         st.sendResponse(response);

      } catch (Exception ex) {
         ex.printStackTrace();
         logger.error("Exiting.");
         System.exit(0);
      }
   }

   /** Process the bye request.
   ----------------------------------------------------------------
    */
   private void processBye( RequestEvent requestEvent, ServerTransaction serverTransactionId) {
      SipProvider sipProvider;
      Request     request;
      Response    response;
      
      sipProvider = (SipProvider) requestEvent.getSource();
      request     = requestEvent.getRequest();
      try {
         response = messageFactory.createResponse(Response.OK, request);
         serverTransactionId.sendResponse(response);
      } catch (Exception ex) {
         ex.printStackTrace();
         //System.exit(0);
      }
   }

   private void subscribeToPeerList(String peerIPAddress) {

      SipURI            peerListURI;
      SipURI            requestURI;
      Request           request;
      Address           address;
      ToHeader          toHeader;
      ViaHeader         viaHeader;
      CallIdHeader      callIdHeader;
      CSeqHeader        cSeqHeader;
      MaxForwardsHeader maxForwardsHeader;
      EventHeader       eventHeader;
      AcceptHeader      acceptHeader;
      ClientTransaction clientTransID;
      ArrayList         viaHeaders;
            
      // why do I need these 3 loc again??  Sat Sep  4 16:31:54 PDT 2004
      if( subscriptions.isSubscription(peerIPAddress) || subscriptions.isPending(peerIPAddress) ) 
         return;
      subscriptions.addPending(peerIPAddress);
      
      try {
         peerListURI = addressFactory.createSipURI("peer-list", peerIPAddress);
         requestURI  = (SipURI )peerListURI.clone();
         requestURI.setTransportParam(transportProto);

         address     = addressFactory.createAddress("Peer List", peerListURI);
         toHeader    = headerFactory.createToHeader(address, null);
         myFromHeader.setTag(Utils.generateTag());

         viaHeaders  = new ArrayList();
         viaHeader   = headerFactory.createViaHeader(
                          sipStack.getIPAddress(),
                          Integer.parseInt(transportPort),
                          transportProto,
                          null
                       );
         viaHeaders.add(viaHeader);

         callIdHeader      = sipProvider.getNewCallId();
         cSeqHeader        = headerFactory.createCSeqHeader(1, Request.SUBSCRIBE);
         maxForwardsHeader = headerFactory.createMaxForwardsHeader(70);
         eventHeader       = headerFactory.createEventHeader("x-earthlink-sip-fpx");
         acceptHeader      = headerFactory.createAcceptHeader("application", peerListMIMEType);

         request           = messageFactory.createRequest(
                                requestURI,
                                Request.SUBSCRIBE,
                                callIdHeader,
                                cSeqHeader,
                                myFromHeader,
                                toHeader,
                                viaHeaders,
                                maxForwardsHeader
                             );

         request.addHeader(myContactHeader);
         request.addHeader(eventHeader);
         request.addHeader(acceptHeader);

         clientTransID = sipProvider.getNewClientTransaction(request);

         clientTransID.sendRequest();
      } catch (Exception e) {
         e.printStackTrace();
         logger.error("Exiting.");
         System.exit(0);
      }
   }

   private void sendPeerListNotify(Subscriber subscriber) {
      int                     cseq;
      int                     expiresIn;
      long                    expiresAt;
      SipURI                  requestURI;
      Dialog                  dialog;
      Request                 request;
      ContentTypeHeader       contentTypeHeader;
      EventHeader             eventHeader;
      SubscriptionStateHeader subscriptionStateHeader;
      ClientTransaction       clientTransID;
         
      try {
         dialog                  = subscriber.getDialog();

         eventHeader             = headerFactory.createEventHeader("x-earthlink-sip-fpx");
         contentTypeHeader       = headerFactory.createContentTypeHeader("application", peerListMIMEType);
         subscriptionStateHeader = headerFactory.createSubscriptionStateHeader("active");
         subscriptionStateHeader.setExpires(oneDay); 
         
         request                 = dialog.createRequest(Request.NOTIFY);
        
         request.setHeader(myContactHeader);
         request.setHeader(eventHeader);
         request.setHeader(subscriptionStateHeader);
         request.setContent(knownHosts.toXML(), contentTypeHeader);
         
         logger.debug("peerList Notify request = " + request);
        
         clientTransID = sipProvider.getNewClientTransaction(request);
       
         dialog.sendRequest(clientTransID);
      } catch (Exception e) {
         e.printStackTrace();
         logger.error("Exiting.");
         System.exit(0);
      }

   }

   private void sendSearchNotify(Request origRequest) {
      byte[]                  messageBody;
      Document                doc;
      Element                 element;
      Iterator                descendants;
      String                  regex                  = null;
      String                  hitList                = null;
      int                     cseq;

      SipURI                  requestURI;
      SipURI                  toURI;
      CallIdHeader            callIdHeader;
      Request                 request;
      Response                response;
      ToHeader                toHeader;
      CSeqHeader              cSeqHeader;
      ContentTypeHeader       contentTypeHeader;
      MaxForwardsHeader       maxForwardsHeader;
      EventHeader             eventHeader;
      SubscriptionStateHeader subscriptionStateHeader;
      ContactHeader           contactHeader;
      ViaHeader               viaHeader;
      ArrayList               viaHeaders;
         
      // Get the regex on which to search our file cache...

      messageBody = origRequest.getRawContent();
      try {
         doc         = new SAXBuilder().build(new ByteArrayInputStream(messageBody));
         descendants = doc.getRootElement().getDescendants();
         while( descendants.hasNext() ) {
            Object o = descendants.next();
            if( o instanceof Element ) {
               element = (Element )o;
               if( element.getName().equals("regex") ) {
                  regex = element.getText();
                  break;
               }
            }
         }
      }
      catch (JDOMException e) {
         logger.error("JDOMException parsing filter document.  Request = " + origRequest);
      }
      catch (IOException e) {
         logger.error("IOException parsing filter document.  Request = " + origRequest);
      }

      if( regex == null )
         return;

      // Generate the list of cache hits based on the request regex...

      hitList = shareList.hitsToXML(regex);
      logger.debug("Cache hits = " + hitList);

      // ... and send the resulting hit list to the Contact field

      try {

         // Send the Notify to whoever is in the Contact: field
         contactHeader     = (ContactHeader)(origRequest.getHeader(ContactHeader.NAME));
         toURI             = (SipURI ) ( contactHeader.getAddress() ).getURI();
         toHeader          = headerFactory.createToHeader(addressFactory.createAddress(toURI), null);

         // Do not attempt to send to yourself 
         if( myStackAddress.equals(toURI.getHost()) )
            return;

         requestURI        = (SipURI )toURI.clone();

         callIdHeader      = sipProvider.getNewCallId();
         cSeqHeader        = headerFactory.createCSeqHeader(1, Request.NOTIFY);
         maxForwardsHeader = headerFactory.createMaxForwardsHeader(70);

         viaHeaders  = new ArrayList();
         viaHeader   = headerFactory.createViaHeader(
                          sipStack.getIPAddress(),
                          Integer.parseInt(transportPort),
                          transportProto,
                          null
                       );
         viaHeaders.add(viaHeader);

         eventHeader             = headerFactory.createEventHeader("x-earthlink-sip-fpx");
         contentTypeHeader       = headerFactory.createContentTypeHeader("application", cacheHitMIMEType);
         subscriptionStateHeader = headerFactory.createSubscriptionStateHeader("terminated");

         request = messageFactory.createRequest(
                      requestURI,
                      Request.NOTIFY,
                      callIdHeader,
                      cSeqHeader,
                      myFromHeader,
                      toHeader,
                      viaHeaders,
                      maxForwardsHeader
                   );

         request.setHeader(myContactHeader);
         request.setHeader(eventHeader);
         request.setHeader(subscriptionStateHeader);
         request.setContent(hitList, contentTypeHeader);
         
         sipProvider.sendRequest(request);

      } catch (Exception e) {
         e.printStackTrace();
         logger.error("Exiting.");
         System.exit(0);
      }

   }

   public void search(String regex) {
      logger.warn("Search request received = " + regex);

      /* Arbitrarily and for the time being, just send search Requests
      to those hosts to which I am subscribed
      */

      for(Iterator i = subscriptions.iterator(); i.hasNext(); ) {
         Subscription sub = (Subscription )i.next();
         subscribeSearch(sub.getHost(), regex);
      }
   }

   private void forward(Request request) {
      SipURI    requestURI;
      SipURI    toURI;
      SipURI    fromURI;
      ToHeader  toHeader;
      ViaHeader viaHeader;
      Address   address;  
      String    host;
      String    fromIP;
      String    thisCallID;

      thisCallID =  ( (CallIdHeader )request.getHeader(CallIdHeader.NAME) ).getCallId();

      // Do not forward messages we have already seen.

      for(Iterator i = messageIDs.iterator(); i.hasNext(); ) {
         String callid;

         callid = (String )i.next();
         if ( thisCallID.equals(callid) ) 
            return;
      }

      fromURI = (SipURI ) ( (FromHeader )(request.getHeader(FromHeader.NAME)) ) .getAddress().getURI();
      fromIP  = fromURI.getHost();

      // Forward to each of my knownHosts

      try {
 
         for(Iterator i = knownHosts.getIterator(); i.hasNext(); ) {

            host    = (String )i.next();

            // Don't forward the request to the sender
            if( ! host.equals(fromIP) ) {

               // Construct a new Request, preserving the original Contact header

               toURI      = addressFactory.createSipURI("search", host);
               address    = addressFactory.createAddress(toURI);
               toHeader   = headerFactory.createToHeader(address, null);

               requestURI = (SipURI )toURI.clone();
   
               viaHeader   = headerFactory.createViaHeader(
                                sipStack.getIPAddress(),
                                Integer.parseInt(transportPort),
                                transportProto,
                                null
                             );
   
               request.setRequestURI(requestURI);
               request.setHeader(toHeader);
               request.setHeader(myFromHeader);
               request.setHeader(viaHeader);
   
               sipProvider.sendRequest(request);
            }

         }
         
         // Mark this request as 'seen'
         messageIDs.add(thisCallID);

      }
      catch (Exception e) {
         e.printStackTrace();
         logger.error("Exiting.");
         System.exit(0);
      }

   }


   private void subscribeSearch(String peerIPAddress, String regex) {
      SipURI            peerURI;
      SipURI            requestURI;
      Request           request;
      Address           address;
      ToHeader          toHeader;
      ViaHeader         viaHeader;
      ContentTypeHeader contentTypeHeader;
      CallIdHeader      callIdHeader;
      CSeqHeader        cSeqHeader;
      MaxForwardsHeader maxForwardsHeader;
      EventHeader       eventHeader;
      AcceptHeader      acceptHeader;
      ArrayList         viaHeaders;
      Document          doc;
      Element           rootElement;
      Element           element;
      String            filterDocument;

      try {
         peerURI = addressFactory.createSipURI("search", peerIPAddress);
         requestURI  = (SipURI )peerURI.clone();
         requestURI.setTransportParam(transportProto);

         address     = addressFactory.createAddress("Search", peerURI);
         toHeader    = headerFactory.createToHeader(address, null);

         viaHeaders  = new ArrayList();
         viaHeader   = headerFactory.createViaHeader(
                          sipStack.getIPAddress(),
                          Integer.parseInt(transportPort),
                          transportProto,
                          null
                       );
         viaHeaders.add(viaHeader);

         callIdHeader      = sipProvider.getNewCallId();
         cSeqHeader        = headerFactory.createCSeqHeader(1, Request.SUBSCRIBE);
         maxForwardsHeader = headerFactory.createMaxForwardsHeader(70);
         eventHeader       = headerFactory.createEventHeader("x-earthlink-sip-fpx");
         acceptHeader      = headerFactory.createAcceptHeader("application", cacheHitMIMEType);
         contentTypeHeader = headerFactory.createContentTypeHeader("application", searchFilterDocumentMIMEType);

         request           = messageFactory.createRequest(
                                requestURI,
                                Request.SUBSCRIBE,
                                callIdHeader,
                                cSeqHeader,
                                myFromHeader,
                                toHeader,
                                viaHeaders,
                                maxForwardsHeader
                             );

         request.addHeader(myContactHeader);
         request.addHeader(eventHeader);
         request.addHeader(acceptHeader);

         rootElement = new Element("search");
         doc         = new Document(rootElement);
         element     = new Element("regex");
         element.addContent(regex);
         rootElement.addContent(element);

         filterDocument = new XMLOutputter(Format.getPrettyFormat()).outputString(doc);

         request.setContent(filterDocument, contentTypeHeader);

         // Register this message as effectively 'seen'
         messageIDs.add(callIdHeader.getCallId());

         sipProvider.sendRequest(request);

         logger.debug("Search request = " + request);

      } catch (Exception e) {
         e.printStackTrace();
         logger.error("Exiting.");
         System.exit(0);
      }
   }

   public void download(Vector v) {            
      SipURI                peerURI;
      SipURI                requestURI;
      Request               request;
      Address               address;
      ToHeader              toHeader;
      ViaHeader             viaHeader;
      ContentTypeHeader     contentTypeHeader;
      CallIdHeader          callIdHeader;
      CSeqHeader            cSeqHeader;
      Dialog                dialog;
      MaxForwardsHeader     maxForwardsHeader;
      EventHeader           eventHeader;
      AllowHeader           allowHeader;
      ClientTransaction     clientTransID;
      DialogApplicationData dialogData;
      ArrayList             viaHeaders;      
      String                peerIPAddress;
      StringBuffer          sdpBody;
      String                fileName;
      String                digest;
      int                   fileSize;

      /* Info:
       v(0) is filename
       v(1) is file size
       v(2) is true digest (requires fspd run in "digest mode")
       v(3) is hosting server's IP address
      */
      
      fileName      = (String )v.get(0);
      fileSize      = Integer.parseInt((String )v.get(1));
      digest        = (String )v.get(2);
      peerIPAddress = (String )v.get(3);
      
      try {
         peerURI = addressFactory.createSipURI(digest, peerIPAddress);         
         requestURI  = (SipURI )peerURI.clone();
         requestURI.setTransportParam(transportProto);

         address     = addressFactory.createAddress("Download", peerURI);
         toHeader    = headerFactory.createToHeader(address, null);
         myFromHeader.setTag(Utils.generateTag());

         viaHeaders  = new ArrayList();
         viaHeader   = headerFactory.createViaHeader(
                          sipStack.getIPAddress(),
                          Integer.parseInt(transportPort),
                          transportProto,
                          null
                       );
         viaHeaders.add(viaHeader);

         callIdHeader      = sipProvider.getNewCallId();
         cSeqHeader        = headerFactory.createCSeqHeader(1, Request.INVITE);
         maxForwardsHeader = headerFactory.createMaxForwardsHeader(70);
         contentTypeHeader = headerFactory.createContentTypeHeader("application", "sdp");         

         request           = messageFactory.createRequest(
                                requestURI,
                                Request.INVITE,
                                callIdHeader,
                                cSeqHeader,
                                myFromHeader,
                                toHeader,
                                viaHeaders,
                                maxForwardsHeader
                             );

         request.addHeader(myContactHeader);
         allowHeader       = headerFactory.createAllowHeader(Request.INVITE);
         request.addHeader(allowHeader);
         //allowHeader       = headerFactory.createAllowHeader(Request.CANCEL);
         //request.addHeader(allowHeader);
         allowHeader       = headerFactory.createAllowHeader(Request.BYE);
         request.addHeader(allowHeader);         
         allowHeader       = headerFactory.createAllowHeader(Request.SUBSCRIBE);
         request.addHeader(allowHeader);
         allowHeader       = headerFactory.createAllowHeader(Request.NOTIFY);
         request.addHeader(allowHeader);
                  
         sdpBody = new StringBuffer();
         sdpBody.append("v=0\n");
         sdpBody.append("o=- 0 0 IN IP4 ").append(peerIPAddress).append("\n");
         sdpBody.append("s=EarthLinkPFX\n");
         sdpBody.append("c=IN IP4 ").append(peerIPAddress).append("\n");
         sdpBody.append("t=0 0\n");
         sdpBody.append("m=data 1960 udp\n");
                  
         request.setContent(sdpBody.toString(), contentTypeHeader);               
         
         clientTransID = sipProvider.getNewClientTransaction(request);
         dialog        = clientTransID.getDialog();

         /* digest-to-fileName binding:
          * 
          * JAIN SIP provides a setApplicationData() interface to associate
          * application-specific data with a dialog.  We Invite-by-digest, and
          * we need a place to save-away the digest-to-filename binding so we 
          * can easily get at the filename when we receive the Invite 200 OK.
          * The setApplicationData() is a natural place to stash
          * the digest-to-filename binding, vs an external hashtable, e.g. 
          */
         
         dialogData = new DialogApplicationData();
         dialog.setApplicationData(dialogData);
         dialogData.setFileName(fileName);
         dialogData.fileSize = fileSize;

         clientTransID.sendRequest();
      } catch (Exception e) {
         e.printStackTrace();
         logger.error("Exiting.");
         System.exit(0);
      }
   }
   
   private void sendBye(Dialog dialog) {      
      Request           byeRequest;
      ClientTransaction trans;
            
      try {
         byeRequest = dialog.createRequest(Request.BYE);
         trans      = sipProvider.getNewClientTransaction(byeRequest);
         dialog.sendRequest(trans);
      }
      catch (SipException e) {
         e.printStackTrace();
      }           
   }      
   
   private void pingSubscriptions(Subscription subscription) {
      SipURI            peerListURI;
      SipURI            requestURI;
      Request           request;
      Address           address;
      ToHeader          toHeader;
      ViaHeader         viaHeader;
      CallIdHeader      callIdHeader;
      CSeqHeader        cSeqHeader;
      MaxForwardsHeader maxForwardsHeader;
      EventHeader       eventHeader;
      AcceptHeader      acceptHeader;
      ClientTransaction clientTransID;
      ArrayList         viaHeaders;
      String            peerIPAddress;

      /* Send a stateful Subscribe, keying off an SIP timeout to determine
      whether to remove this host from the list of subscriptions */

      try {

         peerIPAddress = subscription.getHost();

         peerListURI = addressFactory.createSipURI("ping", peerIPAddress);
         requestURI  = (SipURI )peerListURI.clone();
         requestURI.setTransportParam(transportProto);

         address     = addressFactory.createAddress("Ping", peerListURI);
         toHeader    = headerFactory.createToHeader(address, null);
         myFromHeader.setTag(Utils.generateTag());

         viaHeaders  = new ArrayList();
         viaHeader   = headerFactory.createViaHeader(
                          sipStack.getIPAddress(),
                          Integer.parseInt(transportPort),
                          transportProto,
                          null
                       );
         viaHeaders.add(viaHeader);

         callIdHeader      = sipProvider.getNewCallId();
         cSeqHeader        = headerFactory.createCSeqHeader(1, Request.SUBSCRIBE);
         maxForwardsHeader = headerFactory.createMaxForwardsHeader(70);
         eventHeader       = headerFactory.createEventHeader("x-earthlink-sip-fpx");
         acceptHeader      = headerFactory.createAcceptHeader("application", peerListMIMEType);

         request           = messageFactory.createRequest(
                                requestURI,
                                Request.SUBSCRIBE,
                                callIdHeader,
                                cSeqHeader,
                                myFromHeader,
                                toHeader,
                                viaHeaders,
                                maxForwardsHeader
                             );

         request.addHeader(myContactHeader);
         request.addHeader(eventHeader);

         // No Notify's are expected in return.  Not sure what to put here.
         //request.addHeader(acceptHeader);

         clientTransID = sipProvider.getNewClientTransaction(request);
         clientTransID.sendRequest();

         logger.warn("Heartbeat ping sent to subscription host = " + peerIPAddress);

      } catch (Exception e) {
         e.printStackTrace();
         logger.error("Exiting.");
         System.exit(0);
      }

   }

   public void updateWindowManager() {
      windowManager.updateHosts();
   }

   //------------------------------------------------------------------

   public static void main(String argv[]) {
      new FPeer(argv).networkStart();
   }

}
