/*
Copyright (c) 2004 EarthLink, Inc.
All rights reserved.

FPeer Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions 
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*
 * Created on Mar 29, 2004
 *
 * To change the template for this generated file go to
 * Window - Preferences - Java - Code Generation - Code and Comments
 */
package net.earthlink.research.p2p;

import java.util.Collections;
import java.util.Iterator;
import java.util.Vector;
import java.util.SortedSet;
import java.util.TreeSet;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;


/**
 * @author Mark Petrovic
 *
 * To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Generation - Code and Comments
 */
public class KnownHosts {

   private String        hash           = null;

   public KnownHosts() {
      hash = hostsHash();
   }
   /**
    * 
    */
   public String toString() {
      StringBuffer sb;
      String       host;
      Subscription subscription;

      sb = new StringBuffer("Known hosts = ");
      for (Iterator e = FPeer.fpeer.netherHosts.iterator(); e.hasNext(); ) {
         host = (String )e.next();
         sb.append(host).append(" ");
      }
      for (Iterator e = FPeer.fpeer.subscriptions.iterator(); e.hasNext(); ) {
         subscription = (Subscription )e.next();
         host = subscription .getHost();
         sb.append(host).append(" ");
      }
      return sb.toString();
   }

   /**
    * 
    * @return
    */
   private String hostsHash() {
      Vector       v;
      SortedSet    sortedSet;
      StringBuffer sb;

      v = new Vector();
      for( Iterator e = FPeer.fpeer.netherHosts.iterator(); e.hasNext(); )
         v.add(e.next());
      for( Iterator e = FPeer.fpeer.subscriptions.iterator(); e.hasNext(); )
         v.add( ((Subscription )e.next() ).getHost());

      sortedSet  = Collections.synchronizedSortedSet(new TreeSet(v));
      sb         = new StringBuffer();
      for( Iterator i = sortedSet.iterator(); i.hasNext(); )
         sb.append((String )i.next());
      sb.append(">\n");

      return sb.toString();
   }

   /**
    * 
    * @return
    */
   public boolean hasChanged() {
      boolean changed;
      String  newHash;

      newHash = hostsHash();
      changed = ! hash.equals(newHash);
      hash    = newHash;
      return changed;
   }

   /**
    * 
    * @return
    */
   public String toXML() {
      Document    doc;
      Element     rootElement;
      Element     element;
      int         nPeers;
      int         peerMax = 7;

      rootElement = new Element("peerlist");
      doc         = new Document(rootElement); 
      // a UDP packet cannot encapsulate much more than 6 or so hosts
      nPeers = 0;
      for (Iterator e = FPeer.fpeer.netherHosts.iterator(); e.hasNext() ;) {
         if( ++nPeers < peerMax ) {
            element = new Element("peer");
            element.addContent((String )e.next());
            rootElement.addContent(element);
         }
         else break;
      }
      for (Iterator e = FPeer.fpeer.subscriptions.iterator(); e.hasNext() ;) {
         if( ++nPeers < peerMax ) {
            element = new Element("peer");
            element.addContent( ((Subscription )e.next()).getHost() );
            rootElement.addContent(element);
         }
         else break;
      }
      return new XMLOutputter(Format.getPrettyFormat()).outputString(doc);
   }

   /**
    * 
    * @param addr
    * @return
    */
   public boolean isKnownHost(String addr) {
      return FPeer.fpeer.netherHosts.isNetherHost(addr) || FPeer.fpeer.subscriptions.isSubscription(addr);
   }

   /**
   */
   public Iterator getIterator() {
      Vector       v;
      Subscription s;

      v = new Vector();

      for (Iterator e = FPeer.fpeer.netherHosts.iterator(); e.hasNext() ;) {
         v.add(e.next());
      }  
      for (Iterator e = FPeer.fpeer.subscriptions.iterator(); e.hasNext() ;) {
         s = (Subscription )e.next();
         v.add(s.getHost());
      }  
      return v.iterator();
   }
}
