/*
Copyright (c) 2004 EarthLink, Inc.
All rights reserved.

FPeer Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions 
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/* SDPMessage.java
 * 
 * Created on Apr 29, 2004
 *
 * 
*/
package net.earthlink.research.p2p;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.StringTokenizer;


/**
 * @author Mark Petrovic
 *
 */
public class SDPMessage {
   
   private int    dataPort;   
   private String versionRecord;
   private String ownerRecord;
   private String sessionNameRecord;
   private String sessionInfoRecord;
   private String connectionInfoRecord;
   private String timeRecord;
   private String mediaRecord;   
   
   public SDPMessage(byte[] body) {
      parse(body);
   }
   
   private void parse(byte[] body) {
      ByteArrayInputStream bis;
      InputStreamReader    isr;
      BufferedReader       br;
      String               line;
      StringTokenizer      tokenizer;
      
      bis = new ByteArrayInputStream(body);
      isr = new InputStreamReader(bis);
      br  = new BufferedReader(isr);
      
      try {
         while( (line = br.readLine()) != null) {
            if( line.startsWith("v")) {
               versionRecord = line;              
            }
            else if (line.startsWith("o")) {
               ownerRecord = line;
            }
            else if (line.startsWith("s")) {
               sessionNameRecord = line;
            }
            else if (line.startsWith("i")) {
               sessionInfoRecord = line;              
            }
            else if (line.startsWith("c")) {
               connectionInfoRecord = line;
            }
            else if (line.startsWith("t") ) {
               timeRecord = line;
            }
            else if ( line.startsWith("m")) {
               mediaRecord = line;
            }
         }   
      }
      catch (IOException e) {
         e.printStackTrace();
         return;
      }
      
      if (mediaRecord != null) {
         tokenizer = new StringTokenizer(mediaRecord);
         tokenizer.nextToken();
         dataPort = Integer.parseInt(tokenizer.nextToken());         
      }
   }

   public int getMediaPort() {
      return dataPort;
   }
   
   public void setMediaPort(int port) {
      StringBuffer sb;
      
      sb = new StringBuffer();
      
      dataPort = port;
      sb.append("m=data ").append(Integer.toString(dataPort)).append(" udp");
      mediaRecord = sb.toString();
   }
   
   public byte[] getBytes() {      
      StringBuffer sb;
      
      sb = new StringBuffer();
      sb.append(versionRecord).append("\r\n");
      sb.append(ownerRecord).append("\r\n");
      sb.append(sessionNameRecord).append("\r\n");      
      sb.append(connectionInfoRecord).append("\r\n");
      sb.append(timeRecord).append("\r\n");
      sb.append(mediaRecord).append("\r\n");
      return new String(sb).getBytes();            
   }
}
