/*
Copyright (c) 2004 EarthLink, Inc.
All rights reserved.

FPeer Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions 
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

// SearchPanel.java

package net.earthlink.research.p2p.ui;

import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import java.awt.BorderLayout;
import java.awt.Point;

import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.ActionEvent;

import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.JTextField;

import javax.swing.table.DefaultTableModel;

import org.apache.log4j.Logger;

import org.jdom.Document;
import org.jdom.Element;

import net.earthlink.research.p2p.FPeer;

/**
*
* @author Mark Petrovic
*/
public class SearchPanel extends JPanel {

   private DefaultTableModel  model;
   private JLabel             statusLabel;
   private ROTable            table;
   private JTextField         textField;
   private JScrollPane        scrollPane;
   private ListSelectionModel selectionModel;
   private DownloadPopup      downloadPopup;
   private Logger             logger;
   private Vector             hostDigestTable;

   public SearchPanel() {

      logger = Logger.getLogger(this.getClass().getName());

      setLayout(new BorderLayout());

      statusLabel    = new JLabel("Searching for...");
      textField      = new JTextField("", 25);
 
      model          = new DefaultTableModel();
      table          = new ROTable(model);
      selectionModel = table.getSelectionModel();
      
      scrollPane     = new JScrollPane(table,
                                    JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
                                    JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
      model.addColumn("File");      
      model.addColumn("Size");
      model.addColumn("Hash");
      model.addColumn("Host");

      add(textField, BorderLayout.NORTH);
      add(statusLabel, BorderLayout.CENTER);
      add(scrollPane, BorderLayout.SOUTH);

      downloadPopup = new DownloadPopup(table);
      
      hostDigestTable = new Vector();

      textField.addActionListener( new ActionListener() {
                                      public void actionPerformed(ActionEvent e) {
                                         final String typedText;
                                                                                  
                                         if( textField.getText().trim().equals("*") ) 
                                            typedText = ".*";
                                         else 
                                            typedText = textField.getText().trim();
                                         textField.setText(null);
                                         if( typedText.length() != 0 ) {
                                            clearTable();
                                            statusLabel.setText("Searching for " + typedText);
                                            logger.warn("User requested regex = " + typedText);
                                            new Thread() {
                                               public void run() {
                                                  setName("UISearchThread");
                                                  FPeer.fpeer.search(typedText);
                                               }
                                            }.start();
                                         }
                                      }
                                   }
                                 );
      
      table.addMouseListener(new MouseAdapter() {
         public void mousePressed(MouseEvent e) {
            int    row;
            JTable target;
            Point  point;
            
            if( e.isPopupTrigger()) {
               target = (JTable )e.getComponent();
               point  = e.getPoint();
               row    = target.rowAtPoint(point);
               target.getSelectionModel().setSelectionInterval(row,row);            
               downloadPopup.show(table, e.getX(), e.getY());
            }
         }
         
         public void mouseReleased(MouseEvent e) {
            // this is the Windows button down-up popup trigger
            mousePressed(e);
         }
      });
   }
   
   protected void update(Document doc, String host) {
      Element elem;
      List    files;
      
      files = doc.getRootElement().getChildren("file");
      logger.debug("searchPanel.update(), cache hit count = " + files.size());

      for( Iterator i = files.iterator(); i.hasNext() ; ) {
         String       digest;
         StringBuffer sb;
         boolean      inTable = false;

         elem   = (Element )i.next();
         digest = elem.getChildText("digest");

         sb = new StringBuffer();
         sb.append(host).append(digest);;
         String newEntry = sb.toString();

         // Don't add hits that are already displayed.
         for( Iterator j = hostDigestTable.iterator(); j.hasNext(); ) {
            String t = (String )j.next();
            inTable = inTable || t.equals(newEntry);
            if ( inTable ) 
               break;
         }
         if( ! inTable ) {
            hostDigestTable.add(newEntry);

            Vector v = new Vector();
            v.add(elem.getChildText("name"));
            v.add(elem.getChildText("size"));
            v.add(digest);
            v.add(host);

            model.addRow(v);
         }
      }
 
      // Force the entire table to be redrawn.
      model.fireTableDataChanged();
   }
   
   private void clearTable() {
      int rows = model.getRowCount()-1;
      logger.debug("Current rowcount = " + rows);
      while( rows >= 0 )
         model.removeRow(rows--);
      hostDigestTable.clear();
   }
}
