/*
Copyright (c) 2004 EarthLink, Inc.
All rights reserved.

FPeer Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions 
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package net.earthlink.research.p2p.ui;

import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Toolkit;

import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.JFrame;

import java.util.Observer;
import java.util.Observable;

import org.jdom.Document;

import org.apache.log4j.Logger;

import net.earthlink.research.p2p.SharedFile;
import net.earthlink.research.fsp.FSPclient;

/**
* The WindowManager class is the main UI window.
* <p>
* @author Mark Petrovic
*/
public final class WindowManager extends JFrame implements Observer {

   private   Logger logger;

   private   Container contentPane;
   private   Dimension screenDim;
   protected TabsPanel tabsPanel;

   /**
   * Constructor
   * @param windowTitle window title of the main window
   */
   public WindowManager(String stackName) {

      logger   = Logger.getLogger(this.getClass().getName());

      addWindowListener(new WindowAdapter() {
                           public void windowClosing(WindowEvent e) {
                              setVisible(false);
                              // Probably should do a round of Subscribe/Expires=0 here, but...
                              System.exit(0);
                           }
                        }
                       );

      tabsPanel = new TabsPanel();

      contentPane = getContentPane();
      contentPane.setLayout(new BorderLayout());
      contentPane.add(tabsPanel);
      screenDim = Toolkit.getDefaultToolkit().getScreenSize();
      setJMenuBar(new MenuBar());
      setSize(575,431);
      setLocation(screenDim.width/2 - screenDim.width/8,
                  screenDim.height/2 - screenDim.height/8);
      setTitle(stackName);
      setVisible(true);

      logger.info("windowManager created");
   }

   public void updateHosts() {
      tabsPanel.updateHosts();
      logger.info("windowManager hosts updated");
   }

   public void updateCacheHits(Document doc, String host) {
      logger.info("windowManager cacheHits updating");
      tabsPanel.updateSearch(doc, host);
   }

   public void addSharedFile(SharedFile sf) {
      tabsPanel.addSharedFile(sf);
   }

   public void addDownloadClient(FSPclient client, String fileName, int fileSize) {
      tabsPanel.downloadTab.addEntry(client, fileName, fileSize);
   }

   // fulfill Observer interface contract
   public void update( Observable observable, Object o ) { 
      if ( observable instanceof FSPclient ) {
         tabsPanel.updateDownload(observable, o);
      }
   }
}
